﻿// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (Ms-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.ComponentModel;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.Net;

namespace Delay
{
    /// <summary>
    /// Class that adds the If-Modified-Since header to WebClient and offers a helper method for asynchronous file updates.
    /// </summary>
    internal class IfModifiedSinceWebClient : WebClient
    {
        /// <summary>
        /// Gets or sets the value to use for the If-Modified-Since HTTP header.
        /// </summary>
        public DateTime? IfModifiedSince { get; set; }

        /// <summary>
        /// Returns a System.Net.WebRequest object for the specified resource.
        /// </summary>
        /// <param name="address">A System.Uri that identifies the resource to request.</param>
        /// <returns>A new System.Net.WebRequest object for the specified resource.</returns>
        protected override WebRequest GetWebRequest(Uri address)
        {
            var webRequest = base.GetWebRequest(address);
            var httpWebRequest = webRequest as HttpWebRequest;
            if ((null != httpWebRequest) && IfModifiedSince.HasValue)
            {
                // Push the IfModifiedSince value through to the created request
                httpWebRequest.IfModifiedSince = IfModifiedSince.Value;
            }
            return webRequest;
        }

        /// <summary>
        /// Result codes for the UpdateFileIfNewer method.
        /// </summary>
        public enum UpdateFileIfNewerResult
        {
            /// <summary>
            /// There was an error during the update.
            /// </summary>
            Error,

            /// <summary>
            /// The file was already up to date with the web content.
            /// </summary>
            UpToDate,

            /// <summary>
            /// The file has been updated to match the current web content.
            /// </summary>
            Updated,
        };

        /// <summary>
        /// Updates a file from a web source if and only if the web content is newer.
        /// </summary>
        /// <param name="path">Path to the file.</param>
        /// <param name="uri">Path to the web content.</param>
        /// <param name="onComplete">Optional action called upon completion of the update.</param>
        public static void UpdateFileIfNewer(string path, Uri uri, Action<UpdateFileIfNewerResult> onComplete)
        {
            // Validate parameters
            if (null == path)
            {
                throw new ArgumentNullException("path");
            }
            if (null == uri)
            {
                throw new ArgumentNullException("uri");
            }

            // Create a web client
            using (var client = new IfModifiedSinceWebClient())
            {
                // Update IfModifiedSince if the file already exists
                if (File.Exists(path))
                {
                    client.IfModifiedSince = File.GetLastWriteTimeUtc(path);
                }

                // Initiate an asynchronous download to a temporary file
                var downloadPath = Path.GetTempFileName();
                client.DownloadFileCompleted += new AsyncCompletedEventHandler(UpdateFileIfNewerDownloadFileCompleted);
                client.DownloadFileAsync(uri, downloadPath, new UpdateFileIfNewerState(path, downloadPath, onComplete));
            }
        }

        /// <summary>
        /// Handles the DownloadFileCompleted event of the asynchronous download.
        /// </summary>
        /// <param name="sender">Event source.</param>
        /// <param name="e">Event arguments.</param>
        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification = "Caller is not interested in handling possible file errors.")]
        private static void UpdateFileIfNewerDownloadFileCompleted(object sender, AsyncCompletedEventArgs e)
        {
            // Get state instance
            var state = (UpdateFileIfNewerState)e.UserState;

            // Initialize result
            UpdateFileIfNewerResult result = (null == e.Error) ? UpdateFileIfNewerResult.Updated : UpdateFileIfNewerResult.Error;

            // See if HttpStatusCode.NotModified was returned
            var webException = e.Error as WebException;
            if (null != webException)
            {
                var httpWebResponse = webException.Response as HttpWebResponse;
                if (null != httpWebResponse)
                {
                    if (HttpStatusCode.NotModified == httpWebResponse.StatusCode)
                    {
                        // Update result
                        result = UpdateFileIfNewerResult.UpToDate;
                    }
                }
            }

            // If new content was downloaded, replace the file
            if (UpdateFileIfNewerResult.Updated == result)
            {
                try
                {
                    File.Delete(state.Path);
                    File.Move(state.DownloadPath, state.Path);
                }
                catch
                {
                    // Update result
                    result = UpdateFileIfNewerResult.Error;
                }
            }

            // Delete the temporary file
            File.Delete(state.DownloadPath);

            // Call the user's completion handler (if provided)
            if (null != state.OnCompleted)
            {
                state.OnCompleted(result);
            }
        }

        /// <summary>
        /// Class that stores state for the asynchronous download.
        /// </summary>
        private class UpdateFileIfNewerState
        {
            public string Path { get; private set; }
            public string DownloadPath { get; private set; }
            public Action<UpdateFileIfNewerResult> OnCompleted { get; private set; }
            public UpdateFileIfNewerState(string path, string downloadPath, Action<UpdateFileIfNewerResult> onCompleted)
            {
                Path = path;
                DownloadPath = downloadPath;
                OnCompleted = onCompleted;
            }
        }
    }
}
